package hpdos_rdma_offloaded.handler;

import java.net.InetSocketAddress;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Properties;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import com.ibm.darpc.DaRPCClientEndpoint;
import com.ibm.darpc.DaRPCClientGroup;
import com.ibm.darpc.DaRPCStream;

import org.rocksdb.RocksDBException;

import hpdos_rdma_offloaded.lib.Follower;
import hpdos_rdma_offloaded.lib.Packet;
import hpdos_rdma_offloaded.protocol.Request;
import hpdos_rdma_offloaded.protocol.RequestType;
import hpdos_rdma_offloaded.protocol.Response;
import hpdos_rdma_offloaded.protocol.RpcProtocol;

public class NetworkHandler {
    StorageHandler storageHandler;
    private ExecutorService executorService;
    public HashMap<String, DaRPCStream<Request, Response>> streams;
    // public static List<Follower> followers = new ArrayList<>();
    boolean isMaster;

    public NetworkHandler(){}

    public NetworkHandler(StorageHandler storageHandler, Properties properties){
        this.storageHandler = storageHandler;
        this.executorService = Executors.newFixedThreadPool(10);
        this.isMaster = Boolean.valueOf((String)properties.get("app.MASTER"));
    }
    
    public void connectToSals() throws Exception{
        // RpcProtocol rpcProtocol = new RpcProtocol();
        // DaRPCClientGroup<Request, Response> group = DaRPCClientGroup.createClientGroup(rpcProtocol, 100, 0, 16, 16); 
        // InetSocketAddress address = new InetSocketAddress("192.168.200.50", 1919);
        // DaRPCClientEndpoint<Request, Response> clientEp = group.createEndpoint();
        // clientEp.connect(address, 1000);
        // DaRPCStream<Request, Response> stream = clientEp.createStream();
        
        // if(!stream.isEmpty()){
        //     this.streams.put("192.168.200.50", stream);
        // }
    }

    public void connectToFollower(String uid , String ip) throws Exception {
        Thread.sleep(3000);
        System.out.println("Making rdma connection to follower.");
        Follower follower = new Follower();
        follower.setFollowerId(uid);
        follower.setIpAddress(ip);
        RpcProtocol rpcProtocol = new RpcProtocol();
        DaRPCClientGroup<Request, Response> group = DaRPCClientGroup.createClientGroup(rpcProtocol, 100, 0, 16, 16); 
        InetSocketAddress address = new InetSocketAddress(ip, 1921);
        DaRPCClientEndpoint<Request, Response> clientEp = group.createEndpoint();
        clientEp.connect(address, 1000);
        DaRPCStream<Request, Response> stream = clientEp.createStream();
        follower.setStream(stream);
        ReplicationHandler.followers.add(follower);
    }
    


    // Change the Futre to CompletableFuture to achieve correct asynchronousness
    public void create(Packet packet) throws InterruptedException, ExecutionException,RocksDBException{
        System.out.println("Received create request for key/value: " + new String(packet.getKey()) + "/" + new String(packet.getValue()));
        storageHandler.create(packet.getKey(), packet.getValue());
        if (this.isMaster)
        {
            System.out.println("Starting replication");
            Future<Response> futureReplication = this.executorService.submit(()->{
                Response replicationResponse = new Response();
                // Write code to replicate the data to other nics
                ReplicationHandler.replicateMetadata(packet);
                return replicationResponse;
            }); 
            Response response = futureReplication.get();     
            System.out.println("Replicating complete Ack" + response.getAck());
        }
    }

    public byte[] read(Packet packet) throws RocksDBException,InterruptedException,ExecutionException{
        System.out.println("Received read request for key/value: " + new String(packet.getKey()));
        byte[] value = storageHandler.read(packet.getKey());
        if(value != null)
            System.out.println("Got value "+ new String(value));
        if (this.isMaster)
        {    
            System.out.println("Starting replication");
            Future<Response> futureReplication = this.executorService.submit(()->{
                Response replicationResponse = new Response();
                // Write code to replicate the data to other nics
                ReplicationHandler.replicateMetadata(packet);
                return replicationResponse;
            }); 
            Response response = futureReplication.get();     
            System.out.println("Replicating complete Ack "+ response.getAck());
        }
        return value;
    }

    public void update(Packet packet) throws InterruptedException, ExecutionException,RocksDBException{
        this.storageHandler.update(packet.getKey(), packet.getValue());
        if (isMaster) {
            Response response = new Response();
            Future<Response> futureReplication = this.executorService.submit(()->{
                // Write code to replicate the data to other nics
                Response replicationResponse = new Response();
                // Write code to replicate the data to other nics
                ReplicationHandler.replicateMetadata(packet);
                return replicationResponse;
            });
    
            Future<Boolean> futureInvalidation = this.executorService.submit(()->{
                sendInvalidationRequest(packet);
                return false;
            });
            response = futureReplication.get();
            futureInvalidation.get();
            System.out.println("Replicating complete Ack "+ response.getAck());
        }
        
        // Write code to parse the responses here.
    }

    // To implement delete
    public void delete(Packet packet) throws RocksDBException,InterruptedException,ExecutionException
    {
        this.storageHandler.delete(packet.getKey());
        if (this.isMaster) {
            System.out.println("Starting replication");
            Future<Response> futureReplication = this.executorService.submit(()->{
                Response replicationResponse = new Response();
                // Write code to replicate the data to other nics
                ReplicationHandler.replicateMetadata(packet);
                return replicationResponse;
            }); 
            Response response = futureReplication.get();     
            System.out.println("Replicating complete "+ response.getAck());
        }
    }

    public void sendInvalidationRequest(Packet packet) throws InterruptedException, ExecutionException{
        Set<Callable<Response>> callables = new HashSet<>();
        for (DaRPCStream<Request, Response> stream: streams.values()) {
            callables.add(() -> {
                Request request = new Request();
                Response response = new Response();
                request.setRequestType(RequestType.INVALIDATE);
                request.setKey(null);
                request.setValue(null);
                response = stream.request(request, response, false).get();
                return response;
            });
        }

        List<Future<Response>> futures = this.executorService.invokeAll(callables);

        for (Future<Response> future: futures) {
            Response response;
            response = future.get(); 
            
        }
    }
}
