package hpdos_rdma_offloaded.handler;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.lang.ClassNotFoundException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.Properties;
import java.util.UUID;

public class RegistrationHandler {
    public ServerSocket server;
    public int port = 9876;
    boolean isMaster;
    int follower_registration_port;
    int sal_registration_port;
    
    public RegistrationHandler(Properties properties) throws IOException {
        this.isMaster = Boolean.valueOf((String)properties.get("app.MASTER"));
        

        if (this.isMaster) {
            this.follower_registration_port = Integer.valueOf((String)properties.get("app.follower_registration_port"));
            this.sal_registration_port = Integer.valueOf((String)properties.get("app.follower_registration_port"));
            startFollowerRegistrationHandlerServer();
            startSalRegistrationHandlerServer();    
        } else {
            // Follower follower = new Follower();
            // follower.setFollowerId( UUID.randomUUID().toString());
            // follower.setIpAddress("192.168.200.20");
            String followerUUID = UUID.randomUUID().toString();
            String followerIp = "192.168.200.20";
            String message = followerUUID + ";" + followerIp;
            sendRegistrationRequest(message);
        }
    }

    private void sendRegistrationRequest(String message) throws IOException {
        System.out.println("Sending a registration request");
        InetAddress host = InetAddress.getByName("192.168.200.20");
        Socket socket = null;
        ObjectOutputStream oos = null;
        socket = new Socket(host.getHostName(), 9876);
        oos = new ObjectOutputStream(socket.getOutputStream());
        oos.writeObject(message);
    }

    public void startFollowerRegistrationHandlerServer(){
        Runnable serverTask = new Runnable() {
            @Override
            public void run() {
                try {
                    InetAddress addr = InetAddress.getByName("192.168.200.20");
                    ServerSocket serverSocket = new ServerSocket(9876, 50, addr);
                    System.out.println("Started the follower registration service...");
                    while (true) {
                        Socket socket = serverSocket.accept();
                        System.out.println("Accepted connection");
                        ObjectInputStream ois = new ObjectInputStream(socket.getInputStream());
                        String message = (String) ois.readObject();
                        String arr[] = message.split(";");
                        String followerUUID = arr[0];
                        String followerIp = arr[1];
                        System.out.println("Got follower ip: " + followerIp + " , with UUID of: " + followerUUID);
                        
                        NetworkHandler networkHandler = new NetworkHandler();
                        networkHandler.connectToFollower(followerUUID, followerIp);
                        System.out.println("RDMA connection establised");
                        // Add the follower details to the respective class/component
                        // clientProcessingPool.submit(new ClientTask(clientSocket));
                    }
                } catch (Exception e) {
                    System.err.println("Unable to process client request");
                    e.printStackTrace();
                }
            }
        };
        Thread serverThread = new Thread(serverTask);
        serverThread.start();
    }

    public void startSalRegistrationHandlerServer(){
        Runnable serverTask = new Runnable() {
            @Override
            public void run() {
                try {
                    ServerSocket serverSocket = new ServerSocket(9875);
                    System.out.println("Started the SAL registration service...");
                    while (true) {
                        Socket socket = serverSocket.accept();
                        ObjectInputStream ois = new ObjectInputStream(socket.getInputStream());
                        String message = (String) ois.readObject();
                        // Send a response back with the list of followers to the SAL
                        // clientProcessingPool.submit(new ClientTask(clientSocket));
                    }
                } catch (IOException | ClassNotFoundException e) {
                    System.err.println("Unable to process client request");
                    e.printStackTrace();
                }
            }
        };
        Thread serverThread = new Thread(serverTask);
        serverThread.start();
    }
}
