package hpdos_rdma_offloaded.service;

import java.io.IOException;
import java.util.concurrent.ExecutionException;

import com.ibm.darpc.DaRPCServerEndpoint;
import com.ibm.darpc.DaRPCServerEvent;
import com.ibm.darpc.DaRPCService;

import org.rocksdb.RocksDBException;

import hpdos_rdma_offloaded.handler.NetworkHandler;
import hpdos_rdma_offloaded.lib.Packet;
import hpdos_rdma_offloaded.protocol.AckType;
import hpdos_rdma_offloaded.protocol.Request;
import hpdos_rdma_offloaded.protocol.RequestType;
import hpdos_rdma_offloaded.protocol.Response;
import hpdos_rdma_offloaded.protocol.RpcProtocol;

public class MetadataServerService extends RpcProtocol implements DaRPCService<Request, Response>{
    NetworkHandler networkHandler;

    public MetadataServerService(NetworkHandler networkHandler){
        this.networkHandler = networkHandler;
    }

    @Override
    public void processServerEvent(DaRPCServerEvent<Request, Response> event) throws IOException {
        Request request = event.getReceiveMessage();
        Response response = event.getSendMessage();
        if (request.getRequestType() == RequestType.PUT) {
            // Write code for create
           Packet packet =  new Packet(request.getRequestType(), request.getKey(), request.getValue());
           try
           {
                this.networkHandler.create(packet);
                response.setAck(AckType.SUCCESS);
                response.setKey(null);
                response.setValue(null);
            }
            catch(RocksDBException ex)
            {
                ex.printStackTrace();
                response.setAck(AckType.DBFAILED);
            }
            catch (InterruptedException | ExecutionException e) {
                e.printStackTrace();
            }
        } 
        else if(request.getRequestType() == RequestType.GET) {
            Packet packet =  new Packet(request.getRequestType(), request.getKey(), null);
            try
            {
                byte[] value = this.networkHandler.read(packet);
                if(value == null)
                {
                    response.setAck(AckType.NOTFOUND);
                }
                else
                {
                    response.setAck(AckType.SUCCESS_WITH_VALUE);
                    response.setValue(value);
                }
            }
            catch(RocksDBException ex)
            {
                ex.printStackTrace();
                response.setAck(AckType.DBFAILED);
            }
            catch (InterruptedException | ExecutionException e) {
                e.printStackTrace();
            }
            // Write code for read
        }   
        else if(request.getRequestType() == RequestType.DELETE){
            try{
                Packet packet =  new Packet(request.getRequestType(), request.getKey(), null);
                this.networkHandler.delete(packet);
                response.setAck(AckType.SUCCESS);
            }
            catch(RocksDBException ex)
            {
                ex.printStackTrace();
                response.setAck(AckType.DBFAILED);
            }catch (InterruptedException | ExecutionException e) {
                e.printStackTrace();
            }
            // Write code for delete
        }
        event.triggerResponse();
    }

    @Override
    public void open(DaRPCServerEndpoint<Request, Response> rpcClientEndpoint) {
        System.out.println("Received new connection...");
    }

    @Override
    public void close(DaRPCServerEndpoint<Request, Response> rpcClientEndpoint) {
        System.out.println("Closing Connection");   
    }
    
}
